# -*- coding: utf-8 -*-

import logging
import re

from unittest.util import unorderable_list_difference

from odoo import api, fields, models, SUPERUSER_ID, _
from odoo.addons.saas_worker.compat import request
from odoo.tools import email_re

from odoo.addons.auth_signup.models.res_partner import SignupError

_logger = logging.getLogger(__name__)


class Users(models.Model):
    _inherit = 'res.users'

    # Speed up JOINs between res_users and res_partner, if ORDER BY only uses res_partner columns
    # it can use indexes (current _order in 13.0 uses "name, login", so indexes aren't effective)
    _order = "name"

    github_login = fields.Char(string="Github Login", compute="_compute_github_login")

    # need to avoid collision with inherited employee fields comimg from hr (diff semantics)
    hr_department_id = fields.Many2one("hr.department", compute="_compute_department", string="Company Department")

    @api.depends()
    def _compute_github_login(self):
        employees = self.env['hr.employee'].search([
            ('user_id', 'in', self.ids),
            ('github_login', '!=', False),
        ])
        gh_login_per_user = {e.user_id.id: e.github_login for e in employees}
        for user in self:
            user.github_login = gh_login_per_user.get(user.id, False)

    # should have been a simple related field, but we need this to work across companies
    @api.depends()
    def _compute_department(self):
        # sudo across companies, as we share departments
        employees = self.env['hr.employee'].sudo().search([('user_id', 'in', self.ids)])
        departments_per_user = {
            employee.user_id.id: employee.department_id
            for employee in employees
        }
        for user in self:
            user.hr_department_id = departments_per_user.get(user.id, False)

    @api.model
    def _name_search(self, name, domain=None, operator='ilike', **kwargs):
        if not domain and operator == 'ilike' and not (name or '').strip():
            # force non-empty domain on res_partner, to allow PG index scans
            domain = [('name', '!=', False)]
        return super()._name_search(name, domain=domain, operator=operator, **kwargs)

    @api.model
    def _get_default_image(self, partner_type, is_company, parent_id):
        """ Disabled creating avatar for users, for performance issues. """
        return False

    @api.model_create_multi
    def create(self, vals_list):
        for vals in vals_list:
            if vals.get('login'):
                vals['login'] = vals['login'].lower()
        return super(Users, self).create(vals_list)

    @api.model
    def _get_login_domain(self, login):
        # make sure login attempts use lowercase logins, as we force user creation
        # in lowercase above. Without this, a mismatch is possible, and logins can fail.
        # Mixed-case/Uppercase logins are not supported, we want to
        # treat logins as case-insensitive, otherwise users get confused all the time.
        return super()._get_login_domain(login.lower() if login else login)

    @api.model
    def _signup_create_user(self, values):
        # signed up users must enter an email address as login
        if not email_re.match(values.get('login') or ''):
            raise SignupError("Login must be a valid email address: %s" % values['login'])
        return super()._signup_create_user(values)

    def write(self, values):
        me = f'{self.env.user.display_name!r} (#{self.env.user.id})'
        if (self.env.user.id == SUPERUSER_ID or self.env.su) and request:
            user = request.env.user or self.browse(request.session.uid).exists()
            if user:
                me = f'{self.env.user.display_name!r} (#{self.env.user.id}) [sudoed]'
            else:
                me = 'Unconnected user in route auth=none [sudoed]'

        groups_before = {u.id: u.groups_id.mapped(lambda g: (g.id, g.display_name)) for u in self}
        companies_before = {u.id: u.company_ids.mapped(lambda c: (c.id, c.display_name)) for u in self}
        company_before = {u.id: u.company_id for u in self}

        if {'login', 'password'}.intersection(values):
            # Set modified user as 'to sync' for s2s auth
            self.env['openerp.enterprise.database.user'].sudo()\
                .search([('user_id', 'in', self.ids), ('to_sync', '=', False)])\
                .write({'to_sync': True})

        r = super(Users, self).write(values)

        groups_current = {u.id: u.groups_id.mapped(lambda g: (g.id, g.display_name)) for u in self}
        companies_current = {u.id: u.company_ids.mapped(lambda c: (c.id, c.display_name)) for u in self}
        company_current = {u.id: u.company_id for u in self}

        def acl_log(action, msg):
            Log = self.env['access.control.log'].sudo()
            Log.create({
                'ip': request and request.httprequest.environ['REMOTE_ADDR'] or '-',
                'user_id': request and request.env.user.id or self.env.uid,
                'target_user_id': u.id,
                'reason': self.env.context.get('update_reason', '-'),
                'action': action,
                'details': msg,
            })

        for u in self:
            diff = unorderable_list_difference(groups_before[u.id], groups_current[u.id])
            if diff != ([], []):
                diff = '- %r\n+ %r' % diff
                fu = f'{u.display_name!r} (#{u.id})'
                msg = "User %s changed the groups of user %s:\n%s" % (me, fu, diff)
                _logger.info(msg)
                acl_log('group_update', msg.replace('\n', '\n\n'))

            diff = unorderable_list_difference(companies_before[u.id], companies_current[u.id])
            if diff != ([], []):
                diff = '- %r\n+ %r' % diff
                fu = f'{u.display_name!r} (#{u.id})'
                msg = "User %s changed the companies of user %s:\n%s" % (me, fu, diff)
                _logger.info(msg)
                acl_log('company_update', msg.replace('\n', '\n\n'))

            if company_before[u.id] != company_current[u.id]:
                acl_log('company_update', "Default company updated from %s [%d] to %s [%d]" % (
                    company_before[u.id].name, company_before[u.id].id,
                    company_current[u.id].name, company_current[u.id].id
                ))

        return r

    def unlink(self):
        domains = self.env['openerp.enterprise.domain'].sudo().search([('user_id', 'in', self.ids)])
        for domain in domains:
            body = _('User (#%s) removed from database.', domain.user_id.id)
            domain.with_context(mail_create_nosubscribe=True).message_post(
                body=body, message_type='comment', subtype_xmlid='mail.mt_note')
        return super().unlink()

    def _mfa_type(self):
        r = super()._mfa_type()
        if r is not None:
            return r
        ICP = self.env['ir.config_parameter'].sudo()
        regex_mfa_required = ICP.get_param("openerp_enterprise.login_regex_mfa_required")
        if regex_mfa_required:
            user = self.sudo()
            if re.match(regex_mfa_required, user.login):
                return 'totp_mail'

        # Employees with a login that's older than this many days will be required
        # to perform mail-based 2FA. This gives a grace period for new joiners, and a
        # way to progressively enable the option for the whole company.
        # Note: at this point self is the *impostor*, if the login is being spoofed.
        employee_grace_days = int(ICP.get_param("openerp_enterprise.employee_mfa_required_after", -1))
        if self.has_group('base.group_user'):
            user_age = (fields.Datetime.now() - self.sudo().create_date).days
            if user_age > (employee_grace_days or -1):
                return 'totp_mail'

    @classmethod
    def _login(cls, db, login, password, user_agent_env):
        user_id = super(Users, cls)._login(db, login, password, user_agent_env)
        if request:
            ip = request.httprequest.environ['REMOTE_ADDR']
            country = request.geoip.country_name or '?'
            status = "Successful" if user_id else "Failed"
            _logger.info(status + ' login for <%s> on db <%s> from %s (%s)',
                         login, db, ip, country)

        return user_id

    def has_enterprise(self):
        self.ensure_one()
        partner_ids = [self.partner_id.id, self.partner_id.commercial_partner_id.id]
        contracts = self.env['sale.order'].sudo().search([
            ('is_subscription', '=', True),
            ('enterprise_has_download', '=', True),
            '|', ('partner_id', 'in', partner_ids),
                 ('enterprise_final_customer_id', 'in', partner_ids),
        ])
        return self.has_group('base.group_user') or bool(contracts)

    def _generate_tutorial_articles(self):
        tuto_article = self.env.ref('openerp_enterprise.tutorial_article', raise_if_not_found=False)
        if tuto_article:
            favorites_to_create = [
                {
                    'sequence': 0,
                    'user_id': user.id,
                    'article_id': tuto_article.id
                }
                for user in self
            ]
            self.env['knowledge.article.favorite'].sudo().create(favorites_to_create)

    def _should_alert_new_device(self):
        res = super()._should_alert_new_device()
        if request and request.session.impostor_info:
            # impersonification with employee magic password
            # self.env.user = user impersonified
            user = self.env['res.users'].sudo().browse(request.session.impostor_info['id'])
            key = request.httprequest.cookies.get('td_id')
            if key:
                if self.env['auth_totp.device']._check_credentials_for_uid(
                    scope="browser", key=key, uid=user.id
                ):  # the device is known -> don't warn
                    return False
            # impersonification from a new device or with wrong td_id (or without 2FA?)
            user._alert_new_device()  # send email to warn the employee
            return False
        return res
